#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include "PS2Defines.h"
#include "AddRegisters.h"
#include "ctexture.h"
#include "csprite.h"
#include "pad.h"

/*
Fifth example based on the tutorial from Henry S Fortuna...
A wobbly sprite so to say, actually 100 sprites that is being manipulated...

Read the seventeen first lessons from Henry S Fortuna. (http://www.hsfortuna.pwp.blueyonder.co.uk/)
The last lesson you read should be the Rendering Text With In-Built SPS2 Functions lesson.

Have fun...

How does it work ???
The big sprite is made of 100 sprites laid out in a 10x10 sprite array (SP00 to SP99).
This will give us 121 points in a 11x11 grid (the + signs in the figure).
The screen coordinates in the grid are between -200,200 to 200,200.

-200,-200                                                          200,-200
+------+------+------+------+------+------+------+------+------+------+
I SP00 I SP01 I SP02 I SP03 I SP04 I SP05 I SP06 I SP07 I SP08 I SP09 I 
+------+------+------+------+------+------+------+------+------+------+
I SP10 I SP11 I SP12 I SP13 I SP14 I SP15 I SP16 I SP17 I SP18 I SP19 I 
+------+------+------+------+------+------+------+------+------+------+
I SP20 I SP21 I SP22 I SP23 I SP24 I SP25 I SP26 I SP27 I SP28 I SP29 I 
+------+------+------+------+------+------+------+------+------+------+
I SP30 I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I 
+------+------+------+------+------+------+------+------+------+------+
I SP40 I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I 
+------+------+------+------+------+------+------+------+------+------+
I SP50 I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I 
+------+------+------+------+------+------+------+------+------+------+
I SP60 I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I 
+------+------+------+------+------+------+------+------+------+------+
I SP70 I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I 
+------+------+------+------+------+------+------+------+------+------+
I SP80 I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I SPXX I 
+------+------+------+------+------+------+------+------+------+------+
I SP90 I SP91 I SP92 I SP93 I SP94 I SP95 I SP96 I SP97 I SP98 I SP99 I 
+------+------+------+------+------+------+------+------+------+------+
-200,200                                                           200,200

Then I move the points around a little and set each corner of the sprite
so it matches the grid...
The function to move the points around are:
Angle += 0.4;
xPosArr[col+row*11] += (int)(cosf(Angle + (float)(col+row*11))*1.3);
yPosArr[col+row*11] += (int)(sinf(Angle + (float)(col+row*11))*1.3);

The Cosine and Sine functions are used since they are cyclic and adding them
to a point will sooner or later take the point back to it's original position...
Try and experiment with ohter values than 1.3 and 0.4 if you want to...

Have fun and let me know if you have any questions...

Written by Bamse 2004...
bamse@playstation2-linux.com
*/

//We are agoing to need 16 K for the sprites and 4K for the DMa tags...
#define PRIMITIVEMEM_16K 	0
#define DMAMEM_4K			(KILOBYTES(16))
#define TOTALMEMNEEDED		(KILOBYTES(20))

// The SPS2 handle.
int iSPS2Desc;

// The DMA registers, tags etc... we will be using.
Dn_CHCR_t chcr;
Dn_MADR_t madr;
sps2DMATag_t DMACnt;
sps2DMATag_t DMARef;
sps2DMATag_t DMAEnd;
uint64 * pDMATag;

// Memory handeling
sps2Memory_t * pMemCached;
char * pMemPosition;
char * pStartMemPosition;

// Number of QWords to transfer.
int iQWC = 0;

// Angle for the wobbling.
float Angle = 0.0;

// Sprite and x,yPos Array.
CSprite * spriteArr[100];
int xPosArr[11*11];
int yPosArr[11*11];

int main(void)
{

	// Create the sprites...
	for (int row = 0; row<10 ; row++) {
		for (int col = 0; col<10 ; col++) {
			// The sprite size and position doesn't matter, we will be setting the corner positions ourself.
			spriteArr[col+row*10] = new CSprite( 0, 0, 0, 0);
			// Set the Texture positions, 0.0, 0.0 is top Left and 1.0, 1.0 is bottom right.
			// Since we are using 100 sprites, these will be in 0.1 steps.
			spriteArr[col+row*10]->setTexturePos(col*0.1, col*0.1+0.1, row*0.1, row*0.1+0.1);

			// Set the positions in the grid.
			// Doing it this way will set most points more than once, but we are already in the loop so to say... ;-)
			// Left upper corner
			xPosArr[col+row*11] = -200+(col*40);
			yPosArr[col+row*11] = -200+(row*40);

			// Right upper corner
			xPosArr[col+row*11 + 1] = -200+(col*40) + 40;
			yPosArr[col+row*11 + 1] = -200+(row*40);

			// Left lower corner
			xPosArr[col+row*11 + 11] = -200+(col*40);
			yPosArr[col+row*11 + 11] = -200+(row*40) + 40;

			// Right lower corner
			xPosArr[col+row*11 + 12] = -200+(col*40) + 40;
			yPosArr[col+row*11 + 12] = -200+(row*40) + 40;
		}
	}

	// Initialise the Pad connected to port 1 of the PS2, we initilize it as analogue so we can read the pressure.
	if(pad_init (PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE ) == 0)
	{
		printf("Unable to initialise the pads\n");
		pad_cleanup (PAD_0);
		exit(0);
	}

	//initialise sps2
	iSPS2Desc = sps2Init();

	// Initialise the screen
	sps2UScreenInit(0);

	// Allocate the SPS2 memory
	pMemCached = sps2Allocate(TOTALMEMNEEDED, SPS2_MAP_BLOCK_4K | SPS2_MAP_CACHED, iSPS2Desc);

	// Get a char pointer, we can use this to access memory addresses
	// relative to the start of SPS2 memory
	pStartMemPosition = (char *)pMemCached->pvStart;

	// The texture class for the bitmap
	CTexture Tex(iSPS2Desc);

	// Load bitmap into SPS2 memory
	if(!Tex.LoadBitmap("sun.bmp", true, false))
	{
		printf("Failed to load texture.");
		return -1;
	}
	
	//upload the game texture
	Tex.Upload(sps2UScreenGetFirstFreeGSPage());

	do
	{
		// Read the pad.
		pad_update (PAD_0);

		// Get first memory position from the allocated memory.
		pMemPosition = (char *) pMemCached->pvStart;

		//clear the screen to the specified colour
		sps2UScreenClear(0x00, 0x00, 0x00);

		// Print title...
		sps2UPrintf(0,"\377808080\376(250,10)Wobbly sprite demo...");
		sps2UPrintfRender();

		// Increase angle for the wobble effect.
		Angle += 0.4;

		// Move points in the grid around a little...
		// Use points 1 to 10 so the sides will stay fixed, if you want the sides to wobble, use 0-11...
		// cosf(Angle + (float)(col+row*11)), the (float)(col+row*11) part is only to give a kind of
		// offset for the wobble effect, need to be constant per grid-point but not constant for the whole grid...
		for (int row = 1; row<10 ; row++) {
			for (int col = 1; col<10 ; col++) {
				xPosArr[col+row*11] += (int)(cosf(Angle + (float)(col+row*11))*1.3);
				yPosArr[col+row*11] += (int)(sinf(Angle + (float)(col+row*11))*1.3);
			}
		}

		// Set points for sprites and then add them to memory with a Z value = 1...
		// We are setting the four corners per sprite in stead of position and rotation etc...
		// This is more flexible, but a little bit more inconvenient...
		for (int row = 0; row<10 ; row++) {
			for (int col = 0; col<10 ; col++) {
				spriteArr[col+row*10]->setPosAll(xPosArr[col+row*11], yPosArr[col+row*11], xPosArr[col+row*11 + 1], yPosArr[col+row*11 + 1],
				xPosArr[col+row*11 + 11], yPosArr[col+row*11 + 11], xPosArr[col+row*11 + 12], yPosArr[col+row*11 + 12] );
				spriteArr[col+row*10]->addSpriteToMem(pMemPosition, 1);
			}
		}

		// Render the sprites...
		// Calculate Quad Words used by the sprites.
		iQWC = ((int)pMemPosition - (int)pStartMemPosition) >> 4;

		// Set the pDMATag to start of DMA memory.
		pDMATag = (uint64 *)(&pStartMemPosition[DMAMEM_4K]);

		// This will add DMA tags until there is no more QWords left.
		// We can only transfer 4K per tag due to the 4K limitation of SPS2.
		int i = 0;
		do {
			DMARef.i64 = 0;
			if (iQWC>256){
				DMARef.s.QWC = 256;	// Transfer no more than 4K at a time
				iQWC = iQWC - 256;
			} else {
				DMARef.s.QWC = iQWC; // Transfer the leftover QWC.
				iQWC = 0;
			}
			DMARef.s.ID = DMA_ID_SOURCE_REF;
			// Get the physical address of each 4K page, and put it into the ref tag.
			DMARef.s.ADDR = sps2GetPhysicalAddress((char *)&pStartMemPosition[PRIMITIVEMEM_16K] + 4096*i, pMemCached);
			// Add the ref tag to the DMA chain.
			*pDMATag = DMARef.i64;
			pDMATag += 2;
			i++;
		} while (iQWC > 0);

	
		// Now that we have uploaded the sprites, the DMA chain needs to be ended.
		DMAEnd.i64 = 0;
		DMAEnd.s.QWC = 0;	
		DMAEnd.s.ID = DMA_ID_SOURCE_END;

		// Add the DMA tag to the chain.
		*pDMATag = DMAEnd.i64;
		pDMATag += 2;

		// Flush the DMA chain to memory before we fire it off.
		sps2FlushCache(iSPS2Desc);

		// Don't transfer any data from MADR
		*EE_D2_QWC = 0;

		// Set the address of the first tag (right at the start of the chain)
		madr.i32 = 0;
		madr.s.ADDR = sps2GetPhysicalAddress(&pStartMemPosition[DMAMEM_4K], pMemCached);
		*EE_D2_TADR = madr.i32;

		// Set the channel control register, we are using chain mode.
		chcr.i32 = 0;
		chcr.s.MOD = CHCR_MOD_CHAIN;
		chcr.s.STR = 1;

		// Important: Make sure all the data is written to the physical memory.
		sps2FlushCache(iSPS2Desc);	

		// Set the register to start the transfer.
		*EE_D2_CHCR = chcr.i32;	

		// Wait for it to complete.
		sps2WaitForDMA(2, iSPS2Desc);

		// Display the buffer we were just drawing to.
		sps2UScreenSwap();

	// Press start and select together to end the loop, this can be done anytime in the game...
	}while(!((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)));
	
	// Shutdown the screen
	sps2UScreenShutdown();

	// Release SPS2
	sps2Release(iSPS2Desc);

	// Release pad.
	pad_cleanup (PAD_0);

	// Release Sprite memory...
	for (int i = 0; i<100 ; i++) {
		delete spriteArr[i];
	}


}


